@@ -1,3 +1,8 @@
+/* Copyright (C) 2012-2015 Mark Nunberg.
+ *
+ * See included LICENSE file for license details.
+ */
+
 #include "jsonsl.h"
 #include <assert.h>
 #include <limits.h>
@@ -85,45 +90,16 @@ case '8': \
 case '9': \
 case '0':
 
-
-
-/**
- * This table (predeclared) contains characters which are recognized
- * non-string values.
- */
-static jsonsl_special_t *Special_table;
-#define extract_special(c) \
-    Special_table[(unsigned int)(c & 0xff)]
-
-/**
- * This table (predeclared) contains the tokens and other characters
- * which signal the termination of the non-string values.
- */
-static int *Special_Endings;
-#define is_special_end(c) \
-    Special_Endings[(unsigned int)c & 0xff]
-
-/**
- * This table contains entries for the allowed whitespace
- * as per RFC 4627
- */
-static int *Allowed_Whitespace;
-#define is_allowed_whitespace(c) \
-    (c == ' ' || Allowed_Whitespace[(unsigned int)c & 0xff])
-
-
-/**
- * This table contains allowed two-character escapes
- * as per the RFC
- */
-static int *Allowed_Escapes;
-#define is_allowed_escape(c) \
-    Allowed_Escapes[(unsigned int)c & 0xff]
+static unsigned extract_special(unsigned);
+static int is_special_end(unsigned);
+static int is_allowed_whitespace(unsigned);
+static int is_allowed_escape(unsigned);
+static char get_escape_equiv(unsigned);
 
 JSONSL_API
 jsonsl_t jsonsl_new(int nlevels)
 {
-    struct jsonsl_st *jsn =
+    struct jsonsl_st *jsn = (struct jsonsl_st *)
             calloc(1, sizeof (*jsn) +
                     ( (nlevels-1) * sizeof (struct jsonsl_state_st) )
             );
@@ -142,6 +118,7 @@ void jsonsl_reset(jsonsl_t jsn)
     jsn->can_insert = 1;
     jsn->pos = 0;
     jsn->level = 0;
+    jsn->stopfl = 0;
     jsn->in_escape = 0;
     jsn->expecting = 0;
 
@@ -216,6 +193,7 @@ jsonsl_feed(jsonsl_t jsn, const jsonsl_char_t *bytes, size_t nbytes)
         } else if (jsn->action_callback) { \
             jsn->action_callback(jsn, JSONSL_ACTION_##action, state, (jsonsl_char_t*)c); \
         } \
+        if (jsn->stopfl) { return; } \
     }
 
     /**
@@ -231,6 +209,15 @@ jsonsl_feed(jsonsl_t jsn, const jsonsl_char_t *bytes, size_t nbytes)
             INVOKE_ERROR(SPECIAL_EXPECTED); \
         }
 
+#define STATE_SPECIAL_LENGTH \
+    (state)->nescapes
+
+#define IS_NORMAL_NUMBER \
+    ((state)->special_flags == JSONSL_SPECIALf_UNSIGNED || \
+        (state)->special_flags == JSONSL_SPECIALf_SIGNED)
+
+#define STATE_NUM_LAST jsn->tok_last
+
     const jsonsl_uchar_t *c = (jsonsl_uchar_t*)bytes;
     size_t levels_max = jsn->levels_max;
     struct jsonsl_state_st *state = jsn->stack + jsn->level;
@@ -238,7 +225,7 @@ jsonsl_feed(jsonsl_t jsn, const jsonsl_char_t *bytes, size_t nbytes)
     jsn->base = bytes;
 
     for (; nbytes; nbytes--, jsn->pos++, c++) {
-        register jsonsl_type_t state_type;
+        unsigned state_type;
         INCR_METRIC(TOTAL);
         /* Special escape handling for some stuff */
         if (jsn->in_escape) {
@@ -279,21 +266,77 @@ jsonsl_feed(jsonsl_t jsn, const jsonsl_char_t *bytes, size_t nbytes)
             INCR_METRIC(STRINGY_SLOWPATH);
 
         } else if (state_type == JSONSL_T_SPECIAL) {
+            /* Fast track for signed/unsigned */
+            if (IS_NORMAL_NUMBER) {
+                if (isdigit(CUR_CHAR)) {
+                    state->nelem = (state->nelem * 10) + (CUR_CHAR-0x30);
+                    goto GT_NEXT;
+                } else {
+                    goto GT_SPECIAL_NUMERIC;
+                }
+
+            } else if (state->special_flags == JSONSL_SPECIALf_DASH) {
+                if (!isdigit(CUR_CHAR)) {
+                    INVOKE_ERROR(INVALID_NUMBER);
+                }
+
+                if (CUR_CHAR == '0') {
+                    state->special_flags = JSONSL_SPECIALf_ZERO|JSONSL_SPECIALf_SIGNED;
+                } else if (isdigit(CUR_CHAR)) {
+                    state->special_flags = JSONSL_SPECIALf_SIGNED;
+                    state->nelem = CUR_CHAR - 0x30;
+                } else {
+                    INVOKE_ERROR(INVALID_NUMBER);
+                }
+
+                goto GT_NEXT;
+
+            } else if (state->special_flags == JSONSL_SPECIALf_ZERO) {
+                if (isdigit(CUR_CHAR)) {
+                    /* Following a zero! */
+                    INVOKE_ERROR(INVALID_NUMBER);
+                }
+                /* Unset the 'zero' flag: */
+                if (state->special_flags & JSONSL_SPECIALf_SIGNED) {
+                    state->special_flags = JSONSL_SPECIALf_SIGNED;
+                } else {
+                    state->special_flags = JSONSL_SPECIALf_UNSIGNED;
+                }
+                goto GT_SPECIAL_NUMERIC;
+            }
+
             if (state->special_flags & JSONSL_SPECIALf_NUMERIC) {
+                GT_SPECIAL_NUMERIC:
                 switch (CUR_CHAR) {
                 CASE_DIGITS
-                    state->nelem = (state->nelem*10) + (CUR_CHAR-0x30);
+                    STATE_NUM_LAST = '1';
+                    goto GT_NEXT;
+
+                case '.':
+                    if (state->special_flags & JSONSL_SPECIALf_FLOAT) {
+                        INVOKE_ERROR(INVALID_NUMBER);
+                    }
+                    state->special_flags |= JSONSL_SPECIALf_FLOAT;
+                    STATE_NUM_LAST = '.';
                     goto GT_NEXT;
 
                 case 'e':
                 case 'E':
+                    if (state->special_flags & JSONSL_SPECIALf_EXPONENT) {
+                        INVOKE_ERROR(INVALID_NUMBER);
+                    }
+                    state->special_flags |= JSONSL_SPECIALf_EXPONENT;
+                    STATE_NUM_LAST = 'e';
+                    goto GT_NEXT;
+
                 case '-':
                 case '+':
-                    state->special_flags |= JSONSL_SPECIALf_EXPONENT;
-                    goto GT_NEXT;
-                case '.':
-                    state->special_flags |= JSONSL_SPECIALf_FLOAT;
+                    if (STATE_NUM_LAST != 'e') {
+                        INVOKE_ERROR(INVALID_NUMBER);
+                    }
+                    STATE_NUM_LAST = '-';
                     goto GT_NEXT;
+
                 default:
                     if (is_special_end(CUR_CHAR)) {
                         goto GT_SPECIAL_POP;
@@ -304,6 +347,8 @@ jsonsl_feed(jsonsl_t jsn, const jsonsl_char_t *bytes, size_t nbytes)
             }
             /* else if (!NUMERIC) */
             if (!is_special_end(CUR_CHAR)) {
+                STATE_SPECIAL_LENGTH++;
+
                 /* Verify TRUE, FALSE, NULL */
                 if (state->special_flags == JSONSL_SPECIALf_TRUE) {
                     VERIFY_SPECIAL("true");
@@ -317,6 +362,34 @@ jsonsl_feed(jsonsl_t jsn, const jsonsl_char_t *bytes, size_t nbytes)
             }
 
             GT_SPECIAL_POP:
+            if (IS_NORMAL_NUMBER) {
+                /* Nothing */
+            } else if (state->special_flags == JSONSL_SPECIALf_ZERO ||
+                    state->special_flags == (JSONSL_SPECIALf_ZERO|JSONSL_SPECIALf_SIGNED)) {
+                /* 0 is unsigned! */
+                state->special_flags = JSONSL_SPECIALf_UNSIGNED;
+            } else if (state->special_flags == JSONSL_SPECIALf_DASH) {
+                /* Still in dash! */
+                INVOKE_ERROR(INVALID_NUMBER);
+            } else if (state->special_flags & JSONSL_SPECIALf_NUMERIC) {
+                /* Check that we're not at the end of a token */
+                if (STATE_NUM_LAST != '1') {
+                    INVOKE_ERROR(INVALID_NUMBER);
+                }
+            } else if (state->special_flags == JSONSL_SPECIALf_TRUE) {
+                if (STATE_SPECIAL_LENGTH != 4) {
+                    INVOKE_ERROR(SPECIAL_INCOMPLETE);
+                }
+                state->nelem = 1;
+            } else if (state->special_flags == JSONSL_SPECIALf_FALSE) {
+                if (STATE_SPECIAL_LENGTH != 5) {
+                    INVOKE_ERROR(SPECIAL_INCOMPLETE);
+                }
+            } else if (state->special_flags == JSONSL_SPECIALf_NULL) {
+                if (STATE_SPECIAL_LENGTH != 4) {
+                    INVOKE_ERROR(SPECIAL_INCOMPLETE);
+                }
+            }
             SPECIAL_POP;
             jsn->expecting = ',';
             if (is_allowed_whitespace(CUR_CHAR)) {
@@ -500,6 +573,8 @@ jsonsl_feed(jsonsl_t jsn, const jsonsl_char_t *bytes, size_t nbytes)
             } else {
                 if (state->type != '{') {
                     INVOKE_ERROR(BRACKET_MISMATCH);
+                } else if (state->nelem && state->nelem % 2 != 0) {
+                    INVOKE_ERROR(VALUE_EXPECTED);
                 }
                 DO_CALLBACK(OBJECT, POP);
             }
@@ -538,9 +613,13 @@ jsonsl_feed(jsonsl_t jsn, const jsonsl_char_t *bytes, size_t nbytes)
                 STACK_PUSH;
                 state->type = JSONSL_T_SPECIAL;
                 state->special_flags = special_flags;
+                STATE_SPECIAL_LENGTH = 1;
+
                 if (special_flags == JSONSL_SPECIALf_UNSIGNED) {
                     state->nelem = CUR_CHAR - 0x30;
+                    STATE_NUM_LAST = '1';
                 } else {
+                    STATE_NUM_LAST = '-';
                     state->nelem = 0;
                 }
                 DO_CALLBACK(SPECIAL, PUSH);
@@ -556,6 +635,9 @@ jsonsl_feed(jsonsl_t jsn, const jsonsl_char_t *bytes, size_t nbytes)
 JSONSL_API
 const char* jsonsl_strerror(jsonsl_error_t err)
 {
+    if (err == JSONSL_ERROR_SUCCESS) {
+        return "SUCCESS";
+    }
 #define X(t) \
     if (err == JSONSL_ERROR_##t) \
         return #t;
@@ -679,19 +761,21 @@ JSONSL_API
 jsonsl_jpr_t
 jsonsl_jpr_new(const char *path, jsonsl_error_t *errp)
 {
-    char *my_copy;
+    char *my_copy = NULL;
     int count, curidx;
-    struct jsonsl_jpr_st *ret;
-    struct jsonsl_jpr_component_st *components;
+    struct jsonsl_jpr_st *ret = NULL;
+    struct jsonsl_jpr_component_st *components = NULL;
     size_t origlen;
     jsonsl_error_t errstacked;
 
+#define JPR_BAIL(err) *errp = err; goto GT_ERROR;
+
     if (errp == NULL) {
         errp = &errstacked;
     }
 
     if (path == NULL || *path != '/') {
-        *errp = JSONSL_ERROR_JPR_NOROOT;
+        JPR_BAIL(JSONSL_ERROR_JPR_NOROOT);
         return NULL;
     }
 
@@ -703,8 +787,7 @@ jsonsl_jpr_new(const char *path, jsonsl_error_t *errp)
             if (*c == '/') {
                 count++;
                 if (*(c+1) == '/') {
-                    *errp = JSONSL_ERROR_JPR_DUPSLASH;
-                    return NULL;
+                    JPR_BAIL(JSONSL_ERROR_JPR_DUPSLASH);
                 }
             }
         }
@@ -713,8 +796,17 @@ jsonsl_jpr_new(const char *path, jsonsl_error_t *errp)
         count++;
     }
 
-    components = malloc(sizeof(*components) * count);
-    my_copy = malloc(strlen(path) + 1);
+    components = (struct jsonsl_jpr_component_st *)
+            malloc(sizeof(*components) * count);
+    if (!components) {
+        JPR_BAIL(JSONSL_ERROR_ENOMEM);
+    }
+
+    my_copy = (char *)malloc(strlen(path) + 1);
+    if (!my_copy) {
+        JPR_BAIL(JSONSL_ERROR_ENOMEM);
+    }
+
     strcpy(my_copy, path);
 
     components[0].ptype = JSONSL_PATH_ROOT;
@@ -733,9 +825,7 @@ jsonsl_jpr_new(const char *path, jsonsl_error_t *errp)
         }
 
         if (pathret == JSONSL_PATH_INVALID) {
-            free(components);
-            free(my_copy);
-            return NULL;
+            JPR_BAIL(JSONSL_ERROR_JPR_BADPATH);
         }
     } else {
         curidx = 1;
@@ -743,15 +833,31 @@ jsonsl_jpr_new(const char *path, jsonsl_error_t *errp)
 
     path--; /*revert path to leading '/' */
     origlen = strlen(path) + 1;
-    ret = malloc(sizeof(*ret));
+    ret = (struct jsonsl_jpr_st *)malloc(sizeof(*ret));
+    if (!ret) {
+        JPR_BAIL(JSONSL_ERROR_ENOMEM);
+    }
+    ret->orig = (char *)malloc(origlen);
+    if (!ret->orig) {
+        JPR_BAIL(JSONSL_ERROR_ENOMEM);
+    }
     ret->components = components;
     ret->ncomponents = curidx;
     ret->basestr = my_copy;
-    ret->orig = malloc(origlen);
     ret->norig = origlen-1;
     strcpy(ret->orig, path);
 
     return ret;
+
+    GT_ERROR:
+    free(my_copy);
+    free(components);
+    if (ret) {
+        free(ret->orig);
+    }
+    free(ret);
+    return NULL;
+#undef JPR_BAIL
 }
 
 void jsonsl_jpr_destroy(jsonsl_jpr_t jpr)
@@ -765,7 +871,7 @@ void jsonsl_jpr_destroy(jsonsl_jpr_t jpr)
 JSONSL_API
 jsonsl_jpr_match_t
 jsonsl_jpr_match(jsonsl_jpr_t jpr,
-                   jsonsl_type_t parent_type,
+                   unsigned int parent_type,
                    unsigned int parent_level,
                    const char *key,
                    size_t nkey)
@@ -797,18 +903,29 @@ jsonsl_jpr_match(jsonsl_jpr_t jpr,
         }
     }
 
-    /* Check numeric array index */
-    if (p_component->ptype == JSONSL_PATH_NUMERIC
-            && parent_type == JSONSL_T_LIST) {
-        if (p_component->idx != nkey) {
-            return JSONSL_MATCH_NOMATCH;
-        } else {
-            if (parent_level == jpr->ncomponents-1) {
-                return JSONSL_MATCH_COMPLETE;
+    /* Check numeric array index. This gets its special block so we can avoid
+     * string comparisons */
+    if (p_component->ptype == JSONSL_PATH_NUMERIC) {
+        if (parent_type == JSONSL_T_LIST) {
+            if (p_component->idx != nkey) {
+                /* Wrong index */
+                return JSONSL_MATCH_NOMATCH;
             } else {
-                return JSONSL_MATCH_POSSIBLE;
+                if (parent_level == jpr->ncomponents-1) {
+                    /* This is the last element of the path */
+                    return JSONSL_MATCH_COMPLETE;
+                } else {
+                    /* Intermediate element */
+                    return JSONSL_MATCH_POSSIBLE;
+                }
             }
+        } else if (p_component->is_arridx) {
+            /* Numeric and an array index (set explicitly by user). But not
+             * a list for a parent */
+            return JSONSL_MATCH_TYPE_MISMATCH;
         }
+    } else if (parent_type == JSONSL_T_LIST) {
+        return JSONSL_MATCH_TYPE_MISMATCH;
     }
 
     /* Check lengths */
@@ -838,9 +955,9 @@ void jsonsl_jpr_match_state_init(jsonsl_t jsn,
     if (njprs == 0) {
         return;
     }
-    jsn->jprs = malloc(sizeof(jsonsl_jpr_t) * njprs);
+    jsn->jprs = (jsonsl_jpr_t *)malloc(sizeof(jsonsl_jpr_t) * njprs);
     jsn->jpr_count = njprs;
-    jsn->jpr_root = calloc(1, sizeof(size_t) * njprs * jsn->levels_max);
+    jsn->jpr_root = (size_t*)calloc(1, sizeof(size_t) * njprs * jsn->levels_max);
     memcpy(jsn->jprs, jprs, sizeof(jsonsl_jpr_t) * njprs);
     /* Set the initial jump table values */
 
@@ -955,11 +1072,6 @@ const char *jsonsl_strmatchtype(jsonsl_jpr_match_t match)
 
 #endif /* JSONSL_WITH_JPR */
 
-/**
- * Maps literal escape sequences with special meaning to their
- * actual control codes (e.g.\n => 0x20)
- */
-static unsigned char *Escape_Maps;
 /**
  * Utility function to convert escape sequences
  */
@@ -968,7 +1080,7 @@ size_t jsonsl_util_unescape_ex(const char *in,
                                char *out,
                                size_t len,
                                const int toEscape[128],
-                               jsonsl_special_t *oflags,
+                               unsigned *oflags,
                                jsonsl_error_t *err,
                                const char **errat)
 {
@@ -1020,9 +1132,10 @@ size_t jsonsl_util_unescape_ex(const char *in,
              * TODO: should the maps actually reflect the desired
              * replacement character in toEscape?
              */
-            if (Escape_Maps[c[1]]) {
+            char esctmp = get_escape_equiv(c[1]);
+            if (esctmp) {
                 /* Check if there is a corresponding replacement */
-                *out = Escape_Maps[c[1]];
+                *out = esctmp;
             } else {
                 /* Just gobble up the 'reverse-solidus' */
                 *out = c[1];
@@ -1090,41 +1203,40 @@ size_t jsonsl_util_unescape_ex(const char *in,
  * This table contains the beginnings of non-string
  * allowable (bareword) values.
  */
-static jsonsl_special_t _special_table[0x100] = {
+static unsigned short Special_Table[0x100] = {
         /* 0x00 */ 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 0x1f */
         /* 0x20 */ 0,0,0,0,0,0,0,0,0,0,0,0,0, /* 0x2c */
-        /* 0x2d */ JSONSL_SPECIALf_SIGNED /* - */, /* 0x2d */
+        /* 0x2d */ JSONSL_SPECIALf_DASH /* <-> */, /* 0x2d */
         /* 0x2e */ 0,0, /* 0x2f */
-        /* 0x30 */ JSONSL_SPECIALf_UNSIGNED /* 0 */, /* 0x30 */
-        /* 0x31 */ JSONSL_SPECIALf_UNSIGNED /* 1 */, /* 0x31 */
-        /* 0x32 */ JSONSL_SPECIALf_UNSIGNED /* 2 */, /* 0x32 */
-        /* 0x33 */ JSONSL_SPECIALf_UNSIGNED /* 3 */, /* 0x33 */
-        /* 0x34 */ JSONSL_SPECIALf_UNSIGNED /* 4 */, /* 0x34 */
-        /* 0x35 */ JSONSL_SPECIALf_UNSIGNED /* 5 */, /* 0x35 */
-        /* 0x36 */ JSONSL_SPECIALf_UNSIGNED /* 6 */, /* 0x36 */
-        /* 0x37 */ JSONSL_SPECIALf_UNSIGNED /* 7 */, /* 0x37 */
-        /* 0x38 */ JSONSL_SPECIALf_UNSIGNED /* 8 */, /* 0x38 */
-        /* 0x39 */ JSONSL_SPECIALf_UNSIGNED /* 9 */, /* 0x39 */
+        /* 0x30 */ JSONSL_SPECIALf_ZERO /* <0> */, /* 0x30 */
+        /* 0x31 */ JSONSL_SPECIALf_UNSIGNED /* <1> */, /* 0x31 */
+        /* 0x32 */ JSONSL_SPECIALf_UNSIGNED /* <2> */, /* 0x32 */
+        /* 0x33 */ JSONSL_SPECIALf_UNSIGNED /* <3> */, /* 0x33 */
+        /* 0x34 */ JSONSL_SPECIALf_UNSIGNED /* <4> */, /* 0x34 */
+        /* 0x35 */ JSONSL_SPECIALf_UNSIGNED /* <5> */, /* 0x35 */
+        /* 0x36 */ JSONSL_SPECIALf_UNSIGNED /* <6> */, /* 0x36 */
+        /* 0x37 */ JSONSL_SPECIALf_UNSIGNED /* <7> */, /* 0x37 */
+        /* 0x38 */ JSONSL_SPECIALf_UNSIGNED /* <8> */, /* 0x38 */
+        /* 0x39 */ JSONSL_SPECIALf_UNSIGNED /* <9> */, /* 0x39 */
         /* 0x3a */ 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 0x59 */
         /* 0x5a */ 0,0,0,0,0,0,0,0,0,0,0,0, /* 0x65 */
-        /* 0x66 */ JSONSL_SPECIALf_FALSE /* f */, /* 0x66 */
+        /* 0x66 */ JSONSL_SPECIALf_FALSE /* <f> */, /* 0x66 */
         /* 0x67 */ 0,0,0,0,0,0,0, /* 0x6d */
-        /* 0x6e */ JSONSL_SPECIALf_NULL /* n */, /* 0x6e */
+        /* 0x6e */ JSONSL_SPECIALf_NULL /* <n> */, /* 0x6e */
         /* 0x6f */ 0,0,0,0,0, /* 0x73 */
-        /* 0x74 */ JSONSL_SPECIALf_TRUE /* t */, /* 0x74 */
+        /* 0x74 */ JSONSL_SPECIALf_TRUE /* <t> */, /* 0x74 */
         /* 0x75 */ 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 0x94 */
         /* 0x95 */ 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 0xb4 */
         /* 0xb5 */ 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 0xd4 */
         /* 0xd5 */ 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 0xf4 */
-        /* 0xf5 */ 0,0,0,0,0,0,0,0,0,0 /* 0xfe */
+        /* 0xf5 */ 0,0,0,0,0,0,0,0,0,0, /* 0xfe */
 };
-static jsonsl_special_t *Special_table = _special_table;
 
 /**
  * Contains characters which signal the termination of any of the 'special' bareword
  * values.
  */
-static int _special_endings[0x100] = {
+static int Special_Endings[0x100] = {
         /* 0x00 */ 0,0,0,0,0,0,0,0,0, /* 0x08 */
         /* 0x09 */ 1 /* <TAB> */, /* 0x09 */
         /* 0x0a */ 1 /* <LF> */, /* 0x0a */
@@ -1152,12 +1264,11 @@ static int _special_endings[0x100] = {
         /* 0xde */ 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 0xfd */
         /* 0xfe */ 0 /* 0xfe */
 };
-static int *Special_Endings = _special_endings;
 
 /**
- * Contains allowable whitespace.
+ * This table contains entries for the allowed whitespace as per RFC 4627
  */
-static int _allowed_whitespace[0x100] = {
+static int Allowed_Whitespace[0x100] = {
         /* 0x00 */ 0,0,0,0,0,0,0,0,0, /* 0x08 */
         /* 0x09 */ 1 /* <TAB> */, /* 0x09 */
         /* 0x0a */ 1 /* <LF> */, /* 0x0a */
@@ -1173,12 +1284,11 @@ static int _allowed_whitespace[0x100] = {
         /* 0xc1 */ 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 0xe0 */
         /* 0xe1 */ 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* 0xfe */
 };
-static int *Allowed_Whitespace = _allowed_whitespace;
 
 /**
  * Allowable two-character 'common' escapes:
  */
-static int _allowed_escapes[0x100] = {
+static int Allowed_Escapes[0x100] = {
         /* 0x00 */ 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 0x1f */
         /* 0x20 */ 0,0, /* 0x21 */
         /* 0x22 */ 1 /* <"> */, /* 0x22 */
@@ -1205,10 +1315,10 @@ static int _allowed_escapes[0x100] = {
         /* 0xf6 */ 0,0,0,0,0,0,0,0,0, /* 0xfe */
 };
 
-static int *Allowed_Escapes = _allowed_escapes;
-
-
-static unsigned char _escape_maps[0x100] = {
+/**
+ * This table contains the _values_ for a given (single) escaped character.
+ */
+static unsigned char Escape_Equivs[0x100] = {
         /* 0x00 */ 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 0x1f */
         /* 0x20 */ 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 0x3f */
         /* 0x40 */ 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 0x5f */
@@ -1229,6 +1339,39 @@ static unsigned char _escape_maps[0x100] = {
         /* 0xf5 */ 0,0,0,0,0,0,0,0,0,0 /* 0xfe */
 };
 
-static unsigned char *Escape_Maps = _escape_maps;
-
+/* Definitions of above-declared static functions */
+static char get_escape_equiv(unsigned c) {
+    return Escape_Equivs[c & 0xff];
+}
+static unsigned extract_special(unsigned c) {
+    return Special_Table[c & 0xff];
+}
+static int is_special_end(unsigned c) {
+    return Special_Endings[c & 0xff];
+}
+static int is_allowed_whitespace(unsigned c) {
+    return c == ' ' || Allowed_Whitespace[c & 0xff];
+}
+static int is_allowed_escape(unsigned c) {
+    return Allowed_Escapes[c & 0xff];
+}
 
+/* Clean up all our macros! */
+#undef INCR_METRIC
+#undef INCR_GENERIC
+#undef INCR_STRINGY_CATCH
+#undef CASE_DIGITS
+#undef INVOKE_ERROR
+#undef STACK_PUSH
+#undef STACK_POP_NOPOS
+#undef STACK_POP
+#undef CALLBACK_AND_POP_NOPOS
+#undef CALLBACK_AND_POP
+#undef SPECIAL_POP
+#undef CUR_CHAR
+#undef DO_CALLBACK
+#undef ENSURE_HVAL
+#undef VERIFY_SPECIAL
+#undef STATE_SPECIAL_LENGTH
+#undef IS_NORMAL_NUMBER
+#undef STATE_NUM_LAST
